#!/usr/bin/perl
use strict;
use mpidali;
use warnings;
use 5.010;
use Getopt::Long qw(GetOptions);

my $DALIDATDIR_1="./DAT/";
my $DALIDATDIR_2="./DAT/";
my $JOBID="test";
my $TITLE="test";
my $NPARA=1;
my $DOWOLF=1;
my $DOPARSI=1;
my $matrix;
my $list1file;
my $list2file;

my $USAGE="Compare each query structure against all target structures. Report structural aligments with a Z-score above 2.\n\nUSAGE: $0 OPTIONS\n\nOPTIONS:\n --query query.list\tlist of structure identifiers [REQUIRED]\n --db target.list\tlist of target identifiers [REQUIRED unless --matrix is set]\n --matrix           \tall-against-all comparison of query structures\n --dat1 PATH       \tpath to directory containing query data\n --dat2 PATH      \tpath to directory containing target data\n --np N          \tnumber of processes if using openmpi\n\nPDB structure data must be imported beforehand using import.pl\n\nOutputs are xxxxX.txt (xxxxX.html) for each query struccture xxxxX.\n--matrix additionally generates files called 'ordered' (similarity matrix), 'newick' and 'newick_unrooted' (dendrograms)\n\n";

GetOptions(
	'query=s' => \$list1file,
	'db=s' => \$list2file,
	'matrix' => \$matrix,
	'dat1=s' => \$DALIDATDIR_1,
	'dat2=s' => \$DALIDATDIR_2,
	'np=i' => \$NPARA,
	'jobid=s' => \$JOBID,
	'title=s' => \$TITLE,
	'wolf=i' => \$DOWOLF,
	'parsi=i' => \$DOPARSI
) or die $USAGE;   

if (!$list1file) { die $USAGE; }
if (!$matrix && !$list2file) { die $USAGE; }
# add / to dalidatpaths if missing
if($DALIDATDIR_1 !~ /\/$/) { $DALIDATDIR_1.='/'; }
if($DALIDATDIR_2 !~ /\/$/) { $DALIDATDIR_2.='/'; }

# lock directory
&lock();

# split list1 to soap/wolf
my $soaplist1='soaplist1';
my $wolflist1='wolflist1';
my(@list)=&split_list1($list1file,$DALIDATDIR_1,$soaplist1,$wolflist1);

if($matrix) { # all against all structure comparisons: generates extra outputs
	&matrix($soaplist1,$wolflist1,$list1file,$NPARA,$DALIDATDIR_1,$DALIDATDIR_1,$JOBID,$TITLE);
} else { # pairwise comparison
	# run soap if needed
	open(INLIST,"<$soaplist1");
	while(<INLIST>) {
		my($cd1)=/(\w+)/;
		system("echo $cd1 > list1 ; cp $list2file list2");
		&wolf('SOAP',$NPARA,$DALIDATDIR_1,$DALIDATDIR_2,0);
	}
	close(INLIST);
	# run wolf & parsi if requested
	open(INLIST,"<$wolflist1");
	while(<INLIST>) {
                my($cd1)=/(\w+)/;
                system("echo $cd1 > list1 ; cp $list2file list2");
		if($DOWOLF > 0 ) { &wolf('WOLF',$NPARA,$DALIDATDIR_1,$DALIDATDIR_2,0); }
		if($DOPARSI > 0 ) { &parsi($NPARA,$DALIDATDIR_1,$DALIDATDIR_2,0); }
		&swaplist12();
		if($DOWOLF > 0 ) { &wolf('WOLF',$NPARA,$DALIDATDIR_2,$DALIDATDIR_1,2); }
	        if($DOPARSI > 0 ) { &parsi($NPARA,$DALIDATDIR_2,$DALIDATDIR_1,2); }
	  	&swaplist12();
	}
	close(INLIST);
	# output HTML
        &FSSP_output($list2file,$DALIDATDIR_1,$DALIDATDIR_2,$JOBID,$TITLE,@list);
}

# unlock directory
&unlock();       
 
exit();

